from typing import Any, Dict, Literal

from pyspark.sql import DataFrame
from pyspark.sql.functions import col, collect_set


def ingest_spark_dataframe(
    spark_dataframe: DataFrame,
    save_mode: Literal["Overwrite", "Append"],
    options: Dict[str, Any],
) -> None:
    """
    Saves a Spark DataFrame in multiple batches based on the 'batch' column values.
    Each batch is partitioned by 'final_group' column and saved according to the provided options.
    Ingest will be performed in parallel according to the `final_group` column values.

    Parameters
    ----------
    spark_dataframe : DataFrame
        DataFrame containing 'batch' and 'grouping' columns.
        The 'batch' column is used to split the data into separate batches.
        The 'final_group' column is used for partitioning each batch.
    save_mode : Literal["append", "overwrite"]
        The Spark `mode` to use when ingesting data.
    options : Dict[str, Any]
        Dictionary of options to configure the DataFrame writer.
        Refer to example for more information.

    Example
    -------
    >>> options = {
            "relationship": "BOUGHT",
            "relationship.save.strategy": "keys",
            "relationship.source.save.mode": "Match",
            "relationship.source.labels": ":Customer",
            "relationship.source.node.keys": "customerID:id",
            "relationship.target.save.mode": "Match",
            "relationship.target.labels": ":Product",
            "relationship.target.node.keys": "customerID:id",
            "relationship.properties": "quantity,order"
        }
    >>> save_batches(my_spark_df, "Append", options)

    Note
    ----
    - The input DataFrame must contain 'batch' and 'grouping' columns
    - Each unique value in the 'batch' column will create a separate save operation
    - Uses the Neo4j Spark Connector for writing data to Neo4j
    """

    assert save_mode in {
        "Append",
        "Overwrite",
    }, "`save_mode` must be either 'Append' or 'Overwrite'"
    assert (
        "batch" in spark_dataframe.columns
    ), "Spark DataFrame must contain column `batch`"
    assert (
        "final_group" in spark_dataframe.columns
    ), "Spark DataFrame must contain column `final_group`"

    batch_list = spark_dataframe.select(collect_set("batch")).first()[0]

    batches = [
        spark_dataframe.filter(col("batch") == batch_value)
        for batch_value in batch_list
    ]

    # write batches serially to Neo4j database
    for batch in batches:
        (
            batch.repartition("final_group")  # define parallel groups for ingest
            .write.mode(save_mode)
            .format("org.neo4j.spark.DataSource")
            .options(**options)
            .save()
        )



from typing import Dict, Tuple

from pyspark.sql import DataFrame, SparkSession


def create_ingest_batches_from_groups(spark_dataframe: DataFrame) -> DataFrame:
    """
    Create batches for ingest into Neo4j.
    Add a `batch` column to the Spark DataFrame identifying which batch the group in that row belongs to.
    Remove `source_group` and `target_group` columns.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.

    Returns
    -------
    DataFrame
        The Spark DataFrame with `batch` column.
    """

    spark: SparkSession = spark_dataframe.sparkSession

    group_count = (
        spark_dataframe.select("source_group")
        .union(spark_dataframe.select("target_group"))
        .distinct()
        .count()
    )

    coloring = color_complete_graph_with_self_loops(group_count)

    coloring_data = [(k[0], k[1], v) for k, v in coloring.items()]

    # Create a DataFrame from the coloring dictionary
    coloring_df = spark.createDataFrame(
        coloring_data, ["source_group", "target_group", "batch"]
    )

    # Join the DataFrames
    result_df = spark_dataframe.join(
        other=coloring_df,
        on=(spark_dataframe.source_group == coloring_df.source_group)
        & (spark_dataframe.target_group == coloring_df.target_group),
        how="left",  # Use left join to keep all records from spark_dataframe
    ).drop(
        coloring_df.source_group,
        coloring_df.target_group,
        spark_dataframe.source_group,
        spark_dataframe.target_group,
    )

    return result_df


def color_complete_graph_with_self_loops(n: int) -> Dict[Tuple[int], int]:
    """
    Colors the edges of a complete graph with self loops using a rotating pattern.

    Parameters
    ----------
    n : int
        Number of vertices in the graph

    Returns
    -------
    Dict[Tuple[int], int]
        Dictionary mapping edges (tuple) to colors (int)
    """

    edge_colors = {}
    current_color = 0

    # Function to add edge with color (handling both orientations)
    def _add_edge_color(v1: int, v2: int, color: int) -> None:
        """Assign a color to an edge - undirected."""
        edge_colors[(min(v1, v2), max(v1, v2))] = color

    def _step_through_edges(v1: int, v2: int, number_of_steps: int, color: int) -> None:
        """
        Assign a color to a group of edges (the number of steps = number of edges) where neither the
        source nor target node are associated with that color already.
        """

        for i in range(number_of_steps):
            vertex1 = (v1 - i) % n
            vertex2 = (v2 + i) % n
            if (min(vertex1, vertex2), max(vertex1, vertex2)) not in edge_colors:
                _add_edge_color(vertex1, vertex2, color)
            # else:
            #     print(f"{min(vertex1, vertex2), max(vertex1, vertex2)} is already colored")

    # even number of nodes
    if n % 2 == 0:
        # Color even-distance edges
        for start in range(n // 2):
            v1 = start
            v2 = start
            _step_through_edges(v1, v2, n // 2 + 1, current_color)
            current_color += 1
        # Color odd-distance edges
        for start in range(n // 2):
            v1 = start
            v2 = start + 1
            _step_through_edges(v1, v2, n // 2, current_color)
            current_color += 1
    # odd number of nodes
    else:
        # color even and odd distance edges
        for start in range(n):
            v1 = start
            v2 = start
            _step_through_edges(v1, v2, n // 2 + 1, current_color)
            current_color += 1

    return edge_colors



from pyspark.sql import DataFrame

from ..utils.grouping import (
    create_final_group_column_from_source_and_target_groups,
    create_value_groupings,
)


def create_node_groupings(
    spark_dataframe: DataFrame,
    source_col: str,
    target_col: str,
    num_groups: int,
) -> DataFrame:
    """
    Create node groupings for parallel ingest into Neo4j.
    Add `source_group`, `target_group` and `final_group` columns to the Spark DataFrame identifying which groups the row belongs in.
    `final_group` is a concatenation of the source and target group values.
    It is recommended to have `num_groups` <= (2 * the number of workers available in the Spark cluster) for monopartite grouping.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.
    source_col : str
        The column indicating the relationship source id.
    target_col : str
        The column indicating the relationship target id.
    num_groups : int
        The desired number of groups to generate. The process may generate less groups as necessary.

    Returns
    -------
    DataFrame
        The Spark DataFrame with added columns `source_group`, `target_group` and `final_group`.
    """

    # stack source and target
    # group by and count
    counts_df = create_value_counts_dataframe(
        spark_dataframe=spark_dataframe, source_col=source_col, target_col=target_col
    )

    keys_sdf = create_value_groupings(
        value_counts_spark_dataframe=counts_df,
        num_groups=num_groups,
        grouping_column="combined_col",
    )

    final_sdf = spark_dataframe.join(
        other=keys_sdf.withColumnRenamed("group", "source_group"),
        on=(spark_dataframe[source_col] == keys_sdf.value),
        how="left",
    ).drop(keys_sdf.value)
    final_sdf = final_sdf.join(
        other=keys_sdf.withColumnRenamed("group", "target_group"),
        on=(spark_dataframe[target_col] == keys_sdf.value),
        how="left",
    ).drop(keys_sdf.value)

    final_sdf = final_sdf.drop("value")

    final_sdf = create_final_group_column_from_source_and_target_groups(final_sdf)

    return final_sdf


def create_value_counts_dataframe(
    spark_dataframe: DataFrame, source_col: str, target_col: str
) -> DataFrame:
    """
    Create a `count` column counting the number of times a value appears in either the `source_col` or `target_col` column.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.
    source_col : str
        The column indicating the relationship source id.
    target_col : str
        The column indicating the relationship target id.
    num_groups : int
        The desired number of groups to generate. The process may generate less groups as necessary.

    Returns
    -------
    DataFrame
        A Spark DataFrame with columns `combined_col` and `count`.
    """

    combined_df = (
        spark_dataframe.select(source_col)
        .withColumnsRenamed({source_col: "combined_col"})
        .unionAll(
            spark_dataframe.select(target_col).withColumnsRenamed(
                {target_col: "combined_col"}
            )
        )
    )

    counts_df = combined_df.groupBy("combined_col").count()

    return counts_df



from pyspark.sql import DataFrame

from ..utils.grouping import create_value_counts_dataframe, create_value_groupings


def create_node_groupings(
    spark_dataframe, partition_col: str, num_groups: int
) -> DataFrame:
    """
    Create node groupings for parallel ingest into Neo4j.
    Add a `final_group` column to the Spark DataFrame identifying which group the row belongs in.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.
    partition_col : str
        The desired column to partition on.
    num_groups : int
        The desired number of groups to generate. The process may generate less groups as necessary.

    Returns
    -------
    DataFrame
        The Spark DataFrame with added column `final_group`.
    """
    ...

    # to create buckets
    # run over partition_col
    # group by and count
    value_counts_sdf = create_value_counts_dataframe(
        spark_dataframe=spark_dataframe, grouping_column=partition_col
    )
    # iterate through the values in max -> min order ex: [{key: Amazon, value_count: 100000}, ...]
    # find most-empty bucket (num_groups) and place value in it and increment bucket value by value_count
    # # track with 2 separate hash maps
    value_groupings_sdf = create_value_groupings(
        value_counts_spark_dataframe=value_counts_sdf,
        num_groups=num_groups,
        grouping_column=partition_col,
    )

    final_sdf = spark_dataframe.join(
        other=value_groupings_sdf.withColumnRenamed("group", "final_group"),
        on=(spark_dataframe[partition_col] == value_groupings_sdf.value),
        how="left",
    ).drop(value_groupings_sdf.value)

    final_sdf = final_sdf.drop("value")

    return final_sdf
